from urllib.parse import urlsplit, SplitResult
from django.conf import settings
import re


class Point(object):
    """
    Describe a point in the space.
    """

    _FLOAT_RE = r'[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?'
    POINT_RE = r'\(?(?P<x>{0}),\s*(?P<y>{0})\)?'.format(_FLOAT_RE)

    @staticmethod
    def from_string(value):
        """
        Convert a string describing a point into a `Point` instance.
        The representation of a point as a string:
            (x, y)
        where `x` and `y` can be signed or unsigned integers or floats
        """
        match = re.match(Point.POINT_RE, value)

        if not match:
            raise ValueError("Value {} is not a valid point".format(value))

        values = match.groupdict()

        return Point(float(values['x']), float(values['y']))

    def __init__(self, x=0, y=0):
        self.x = x
        self.y = y

    def __repr__(self):
        return '<Point({0.x},{0.y})>'.format(self)

    def __str__(self):
        return '({0.x},{0.y})'.format(self)

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self.x == other.x and
                self.y == other.y)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        return (isinstance(other, self.__class__) and
                self.x <= other.x and
                self.y <= other.y)


class BoundingBox(object):
    """
    A bounding box for a list of Point instances or a list of coordinates
    """
    def __init__(self, points):
        assert len(points) > 0
        if not all(isinstance(point, Point) for point in points):
            points = [Point(c[0], c[1]) for c in points]

        all_x = [point.x for point in points]
        all_y = [point.y for point in points]

        self.x, self.y = min(all_x), min(all_y)
        self.width, self.height = max(all_x) - self.x, max(all_y) - self.y

    def __repr__(self):
        return "BoundingBox({}, {}, {}, {})".format(
            self.x, self.y, self.width, self.height)

    def to_polygon(self):
        points = [(self.x, self.y),
                  (self.x, self.y + self.height),
                  (self.x + self.width, self.y + self.height),
                  (self.x + self.width, self.y)]
        return tuple('({},{})'.format(i, j) for i, j in points)


def sslify_url(url):
    """Build a sslified URL if needed."""
    splat = urlsplit(url)

    if settings.SSLIFY_HOST is None or \
            len(settings.SSLIFY_HOST) < 1 or \
            splat.scheme != "http":
        return url

    return SplitResult(
        'https',
        settings.SSLIFY_HOST,
        splat.netloc + splat.path,
        splat.query,
        splat.fragment
    ).geturl()
