from rest_framework import status
from rest_framework.response import Response
from rest_framework.generics import ListAPIView, RetrieveAPIView
from documents.serializers import \
    ElementSerializer, PageLightSerializer, SearchResultSerializer, \
    VolumeManifestSerializer, PageAnnotationListSerializer
from documents.models import Element, ElementType, Page
from documents.search import search_transcriptions


class ElementsList(ListAPIView):
    """
    List documents using pagination
    """
    serializer_class = ElementSerializer

    def get_queryset(self):

        # Apply filters
        documents = Element.objects.filter(
            type=self.request.query_params.get('type', ElementType.Volume)
        )

        return documents


class RelatedElementsList(ListAPIView):
    """
    List all documents linked to a specified document
    """
    serializer_class = ElementSerializer

    def get_queryset(self):
        return Element.objects.filter(
            id__in=Element.objects.get_related(self.kwargs['pk'])
        )


class ElementPages(ListAPIView):
    """
    List pages inside a document
    """
    serializer_class = PageLightSerializer

    def get_queryset(self):
        return Page.objects.get_descending(self.kwargs['pk'])


class TranscriptionSearch(ListAPIView):
    """
    Search and list transcriptions, using pagination
    """
    serializer_class = SearchResultSerializer

    def add_parents(self, results):
        ids = [trans.id for trans in results]
        all_parents = Element.objects.get_ascendings(*ids)
        for trans in results:
            trans.parent_docs = all_parents.get(trans.id, [])
        return results

    def get_queryset(self):
        ''' Django rest framework check this method exists '''

    def list(self, request):
        query = self.request.query_params.get('q')
        if not query:
            return Response(status=status.HTTP_204_NO_CONTENT)
        search_results = search_transcriptions(query)
        if search_results is None:
            return Response(status=status.HTTP_204_NO_CONTENT)

        # Handle pagination, if activated
        page = self.paginate_queryset(search_results)
        if page is not None:
            serializer = self.get_serializer(self.add_parents(page), many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(self.add_parents(search_results), many=True)
        return Response(serializer.data)


class VolumeManifest(RetrieveAPIView):
    """
    Get a IIIF manifest for a specific volume
    """

    serializer_class = VolumeManifestSerializer
    queryset = Element.objects.filter(type=ElementType.Volume)


class PageAnnotationList(RetrieveAPIView):
    """
    Get a IIIF annotation list for a specific page
    """

    serializer_class = PageAnnotationListSerializer
    queryset = Page.objects.all()
