from django.db import models
from django.urls import reverse
from enumfields import EnumField, Enum
from arkindex.models import IndexableModel
from documents.managers import ElementManager
import uuid


class ElementType(Enum):
    Volume = 'volume'
    Register = 'register'
    Act = 'act'
    Page = 'page'
    Transcription = 'transcription'


class ElementLink(models.Model):
    """
    Links elements together through a parent-child relation
    """
    parent = models.ForeignKey('documents.Element', on_delete=models.CASCADE, related_name='link_parent')
    child = models.ForeignKey('documents.Element', on_delete=models.CASCADE, related_name='link_child')
    order = models.PositiveIntegerField(default=0)

    class Meta:
        unique_together = (
            ('parent', 'order'),
            ('parent', 'child'),
        )

    def __str__(self):
        return "'{}' -> '{}'".format(self.parent, self.child)


class ElementParents(models.Model):
    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    child = models.ForeignKey('documents.Element', related_name='ep_child', on_delete=models.DO_NOTHING)
    parent = models.ForeignKey('documents.Element', related_name='ep_parent', on_delete=models.DO_NOTHING)
    ordering = models.PositiveIntegerField()

    class Meta:
        managed = False
        db_table = 'elements_parents'


class ElementChildren(models.Model):
    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    child = models.ForeignKey('documents.Element', related_name='ec_child', on_delete=models.DO_NOTHING)
    parent = models.ForeignKey('documents.Element', related_name='ec_parent', on_delete=models.DO_NOTHING)
    ordering = models.PositiveIntegerField()

    class Meta:
        managed = False
        db_table = 'elements_children'


class Element(IndexableModel):
    """
    aka Unité Codicologique
    """
    parents = models.ManyToManyField('self', through=ElementLink, symmetrical=False)
    type = EnumField(ElementType, max_length=50)
    name = models.CharField(max_length=250)

    objects = ElementManager()

    def build_absolute_url(self, request, name, id_argument='pk', **kwargs):
        """Build an absolute URL for a specified view using the element ID."""
        kwargs[id_argument] = str(self.id)
        return request.build_absolute_uri(reverse(name, kwargs=kwargs))

    def get_thumbnail(self):
        """Get an Image instance corresponding to a thumbnail for this element."""
        if self.type == ElementType.Volume:
            first_page = Page.objects.get_descending(self.id).first()
            return first_page and first_page.images.first()
        elif self.type == ElementType.Page:
            return self.page.images.first()
        return None

    def __str__(self):
        return '{} : {}'.format(self.type, self.name)


class PageType(Enum):
    Page = 'page'
    Summary = 'summary'
    Garde = 'garde'
    ContreGarde = 'contregarde'
    ContrePlat = 'contreplat'
    Plat = 'plat'
    Copy = 'copy'


class PageDirection(Enum):
    Recto = 'recto'
    Verso = 'verso'


class Page(Element):
    """
    with folio numbering
    """
    folio = models.CharField(max_length=250)

    # Parsed folio
    page_type = EnumField(PageType, max_length=50, null=True, blank=True)
    nb = models.PositiveIntegerField(null=True, blank=True)
    direction = EnumField(PageDirection, max_length=10, null=True, blank=True)
    objects = ElementManager()

    @property
    def display_name(self):
        """
        Generate a pretty display name for a page.
        If the folio parsing had failed, will return the raw folio.
        """
        if self.page_type is None and self.nb is None and self.direction is None:
            if self.folio.strip() == "":
                return "Unknown folio"
            return self.folio

        text = self.page_type.name if self.page_type is not None else 'Page'

        if self.nb is not None:
            text += ' no. {0:d}'.format(self.nb)
        else:
            text += ', unnumbered'

        if self.direction is not None:
            text += ', {}'.format(self.direction.value)

        return text.capitalize()

    def save(self, *args, **kwargs):
        # TODO: move this in Element through introspection
        self.type = ElementType.Page
        super().save(*args, **kwargs)

    def __str__(self):
        out = [self.type.value, ]
        if self.nb:
            out.append('n°{}'.format(self.nb))
        if self.direction:
            out.append(self.direction.value)
        return ' '.join(out)


class Section(Element):
    """
    A logical subdivision that can span on multiple images via zones
    """

    def save(self, *args, **kwargs):
        # TODO: move this in Element through introspection
        self.type = ElementType.Act
        super().save(*args, **kwargs)


class Transcription(Element):
    """
    A transcription on:
     * a zone on an image
    """
    INDEX_TYPE = 'transcription'

    # Data
    line = models.PositiveIntegerField(null=True, blank=True)
    text = models.TextField(null=True, blank=True)
    score = models.FloatField(null=True, blank=True)

    def __str__(self):
        return 'Transcription: {}'.format(self.text[:20])

    def save(self, *args, **kwargs):
        # TODO: move this in Element through introspection
        self.type = ElementType.Transcription
        super().save(*args, **kwargs)

    def build_search_index(self):
        """
        Structure indexed into Elastic Search
        """
        return {
            'id': self.id,  # TODO: Could be removed? (can use _id instead)
            'score': self.score,
            'line': self.line,
            'text': self.text,
        }
