from django.test import TestCase
from django.urls import reverse
from rest_framework.test import APITestCase
from rest_framework import status
from tripoli import IIIFValidator
from documents.models import Element, ElementLink, ElementType, Page, Transcription
from images.models import ImageServer, Image, Zone


class TestElementManager(TestCase):
    """Tests for ElementManager class"""

    def setUp(self):
        self.reg = Element.objects.create(name="Register", type=ElementType.Register)
        self.vol = Element.objects.create(name="Volume", type=ElementType.Volume)
        self.p1 = Element.objects.create(name="Page1", type=ElementType.Page)
        self.p2 = Element.objects.create(name="Page2", type=ElementType.Page)
        self.act = Element.objects.create(name="Act", type=ElementType.Act)
        # Register --> Volume --> [Page1 --> Act, Page2]
        ElementLink.objects.create(parent=self.reg, child=self.vol, order=0)
        ElementLink.objects.create(parent=self.vol, child=self.p1, order=0)
        ElementLink.objects.create(parent=self.vol, child=self.p2, order=1)
        ElementLink.objects.create(parent=self.p1, child=self.act, order=0)

    def test_get_ascending(self):
        # Use Page1, expect Register and Volume
        ids = Element.objects.get_ascending(self.p1.id)
        self.assertCountEqual(ids, [self.reg, self.vol])

    def test_get_ascending_empty(self):
        # Use Register, expect empty list
        ids = Element.objects.get_ascending(self.reg.id)
        self.assertCountEqual(ids, [])

    def test_get_ascendings(self):
        # Use Page1 and Page2, expect Register and Volume twice
        ids = Element.objects.get_ascendings(self.p1.id, self.p2.id)
        self.assertCountEqual(ids[self.p1.id], [self.reg, self.vol])
        self.assertCountEqual(ids[self.p2.id], [self.reg, self.vol])

    def test_get_descending(self):
        # Use Volume, expect Page1, Page2 and Act
        ids = Element.objects.get_descending(self.vol.id)
        self.assertCountEqual(ids, [self.p1, self.p2, self.act])

    def test_get_descending_empty(self):
        # Use Page2, expect empty list
        ids = Element.objects.get_descending(self.p2.id)
        self.assertCountEqual(ids, [])

    def test_get_descendings(self):
        # Use Page1 and Page2, expect Act and nothing else
        ids = Element.objects.get_descendings(self.p1.id, self.p2.id)
        self.assertCountEqual(ids[self.p1.id], [self.act])
        self.assertNotIn(self.p2.id, ids)

    def test_get_related(self):
        # Use Page1, expect Register, Volume and Act
        ids = Element.objects.get_related(self.p1.id)
        self.assertCountEqual(ids, [self.reg, self.vol, self.act])

    def test_get_related_empty(self):
        # Use a element linked to nothing, expect empty list
        alone = Element.objects.create(name="LonelyPage", type=ElementType.Page)
        ids = Element.objects.get_related(alone.id)
        self.assertCountEqual(ids, [])


class TestVolumeManifestSerializer(APITestCase):
    """Tests for VolumeManifestSerializer class"""

    def setUp(self):
        # Create a volume with two pages and a server with two images
        self.imgsrv = ImageServer.objects.create(name="Test Server", url="http://server")
        self.img1 = Image.objects.create(path='img1', width=1337, height=42, server=self.imgsrv)
        self.img2 = Image.objects.create(path='img2', width=255, height=420, server=self.imgsrv)
        self.vol = Element.objects.create(name="Volume Name", type=ElementType.Volume)
        self.p1 = Page.objects.create(name="p1", folio="p1")
        self.p2 = Page.objects.create(name="p2", folio="p2")
        Zone.objects.create(polygon=[[0, 0], [1337, 0], [1337, 42], [42, 0]], image=self.img1, element=self.p1)
        Zone.objects.create(polygon=[[0, 0], [255, 0], [255, 420], [0, 420]], image=self.img2, element=self.p2)
        ElementLink.objects.create(parent=self.vol, child=self.p1, order=0)
        ElementLink.objects.create(parent=self.vol, child=self.p2, order=1)

    def test_normal_manifest(self):
        # Just a normal situation.
        response = self.client.get(reverse('api:volume-manifest', kwargs={'pk': self.vol.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        manifest = response.json()

        self.assertIn("@context", manifest)
        self.assertIn("@id", manifest)
        self.assertIn("@type", manifest)
        self.assertIn("thumbnail", manifest)
        self.assertIn("structures", manifest)
        self.assertIn("sequences", manifest)
        self.assertIn("label", manifest)

        self.assertEqual(manifest['@type'], "sc:Manifest")
        self.assertEqual(manifest['@context'],
                         "http://iiif.io/api/presentation/2/context.json")
        self.assertEqual(manifest['label'], self.vol.name)
        self.assertEqual(manifest['thumbnail']['service']['@context'],
                         "http://iiif.io/api/image/2/context.json")
        self.assertEqual(manifest['thumbnail']['service']['profile'],
                         "http://iiif.io/api/image/2/level2.json")
        self.assertIn('server/img', manifest['thumbnail']['service']['@id'])

        self.assertEqual(len(manifest['structures']), 3)
        topstruct = next(s for s in manifest['structures'] if s['viewingHint'] == "top")
        self.assertIsNotNone(topstruct)
        self.assertEqual(topstruct['label'], self.vol.name)
        self.assertEqual(topstruct['@type'], 'sc:Range')
        self.assertEqual(len(topstruct['ranges']), 2)
        self.assertEqual(manifest['@id'], topstruct['@id'])

        self.assertEqual(len(manifest['sequences']), 1)
        self.assertEqual(manifest['sequences'][0]['@type'], 'sc:Sequence')
        canvases = manifest['sequences'][0]['canvases']
        self.assertEqual(len(canvases), 2)
        for canvas in canvases:
            self.assertIn('@type', canvas)
            self.assertIn('@id', canvas)
            self.assertIn('label', canvas)
            self.assertIn('width', canvas)
            self.assertIn('height', canvas)
            self.assertIn('images', canvas)
            self.assertEqual(canvas['@type'], 'sc:Canvas')
            self.assertEqual(len(canvas['images']), 1)

        images = [c['images'][0] for c in canvases]
        for image in images:
            self.assertIn('@type', image)
            self.assertIn('resource', image)
            self.assertIn('on', image)
            self.assertIn('motivation', image)
            self.assertEqual(image['@type'], 'oa:Annotation')
            self.assertEqual(image['motivation'], 'sc:painting')
            res = image['resource']
            self.assertIn('width', res)
            self.assertIn('height', res)
            self.assertIn('service', res)
            self.assertEqual(res['@type'], 'dctypes:Image')
            self.assertEqual(res['service']['@context'],
                             "http://iiif.io/api/image/2/context.json")
            self.assertEqual(res['service']['profile'],
                             "http://iiif.io/api/image/2/level2.json")

    def test_no_page(self):
        # A manifest for an empty volume
        response = self.client.get(reverse('api:volume-manifest', kwargs={
            'pk': Element.objects.create(type=ElementType.Volume, name="Empty Volume").id
        }))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        manifest = response.json()

        self.assertEqual(manifest['thumbnail'], None)
        self.assertEqual(len(manifest['structures']), 1)
        topstruct = manifest['structures'][0]
        self.assertEqual(len(topstruct['ranges']), 0)
        self.assertEqual(len(manifest['sequences']), 1)
        self.assertEqual(len(manifest['sequences'][0]['canvases']), 0)

    def test_normal_manifest_iiif_validation(self):
        response = self.client.get(reverse('api:volume-manifest', kwargs={'pk': self.vol.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        manifest = response.json()
        iv = IIIFValidator()
        iv.validate(manifest)
        self.assertTrue(iv.is_valid)


class TestPageAnnotationListSerializer(APITestCase):

    def setUp(self):
        # Create a page and an image with some transcriptions
        self.imgsrv = ImageServer.objects.create(name="Test Server", url="http://server")
        self.img = Image.objects.create(path='img', width=1337, height=42, server=self.imgsrv)
        self.page = Page.objects.create(name="page", folio="page")
        Zone.objects.create(polygon=[[0, 0], [1337, 0], [1337, 42], [42, 0]],
                            image=self.img, element=self.page)
        self.z1 = Zone.objects.create(polygon=[[100, 200], [100, 300], [300, 300], [300, 200]], image=self.img)
        self.z2 = Zone.objects.create(polygon=[[50, 100], [50, 150], [150, 150], [150, 100]], image=self.img)
        self.t1 = Transcription.objects.create(text="AAA")
        self.t2 = Transcription.objects.create(text="BBB")
        self.t1.zones.add(self.z1)
        self.t2.zones.add(self.z2)
        ElementLink.objects.create(parent=self.page, child=self.t1, order=0)
        ElementLink.objects.create(parent=self.page, child=self.t2, order=1)

    def test_normal_list(self):
        response = self.client.get(reverse('api:page-manifest', kwargs={'pk': self.page.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        annotation_list = response.json()

        self.assertIn('@context', annotation_list)
        self.assertIn('@id', annotation_list)
        self.assertIn('@type', annotation_list)
        self.assertIn('resources', annotation_list)

        self.assertEqual(annotation_list['@type'], 'sc:AnnotationList')
        self.assertEqual(annotation_list['@context'], 'http://iiif.io/api/presentation/2/context.json')
        self.assertEqual(len(annotation_list['resources']), 2)

        for annotation in annotation_list['resources']:
            self.assertIn('@type', annotation)
            self.assertIn('@id', annotation)
            self.assertIn('on', annotation)
            self.assertIn('motivation', annotation)
            self.assertIn('resource', annotation)

            self.assertEqual(annotation['@type'], 'oa:Annotation')
            self.assertEqual(annotation['motivation'], 'sc:painting')

            resource = annotation['resource']
            self.assertIn('@id', resource)
            self.assertIn('@type', resource)
            self.assertIn('format', resource)
            self.assertIn('chars', resource)

            self.assertEqual(resource['format'], 'text/plain')
            self.assertEqual(resource['@type'], 'cnt:ContentAsText')
            self.assertIn(resource['chars'], ['AAA', 'BBB'])

    def test_empty_list(self):
        # An annotation list with nothing in it
        response = self.client.get(reverse('api:page-manifest', kwargs={
            'pk': Page.objects.create(name="Empty Page").id
        }))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        annotation_list = response.json()

        self.assertEqual(len(annotation_list['resources']), 0)
